@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 2000-2014  The R Core Team.
+ *  Copyright (C) 2000-2020  The R Core Team.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, a copy is available at
- *  http://www.r-project.org/Licenses/
+ *  https://www.R-project.org/Licenses/
  *
  *
  *      Interfaces to POSIX date-time conversion functions.
@@ -37,18 +37,26 @@
 # include <config.h>
 #endif
 
-#include <Rmath.h> // Rexp10
+#include <Rmath.h> // for imin2()
 
-// some other header, e.g. math.h, might define it
-#if defined(__GLIBC__) && !defined(_BSD_SOURCE)
-// to get tm_zone, tm_gmtoff defined
-# define _BSD_SOURCE
+// to get tm_zone, tm_gmtoff defined in glibc.
+// some other header, e.g. math.h, might define the macro.
+#if defined HAVE_FEATURES_H
+# include <features.h>
+# ifdef __GNUC_PREREQ
+#  if __GNUC_PREREQ(2,20) && !defined(_DEFAULT_SOURCE_)
+#   define _DEFAULT_SOURCE 1
+#  endif
+# endif
+#endif
+#if defined(HAVE_GLIBC2) && !defined(_DEFAULT_SOURCE_) && !defined(_BSD_SOURCE)
+# define _BSD_SOURCE 1
 #endif
 #include <time.h>
 
 #include <errno.h>
 
-/* 
+/*
 
 There are two implementation paths here.
 
@@ -70,7 +78,7 @@ those limits where there is a 64-bit time_t and the conversions work
 there is code below to extrapolate from 1902-2037.
 
 Path 2) was added for R 3.1.0 and is the only one supported on
-Windows: it is the default on OS X.  The only currently (Jan 2014)
+Windows: it is the default on macOS.  The only currently (Jan 2014)
 known OS with 64-bit time_t and complete tables is Linux.
 
 */
@@ -85,15 +93,13 @@ known OS with 64-bit time_t and complete tables is Linux.
 # define HAVE_TM_GMTOFF 1
 # undef MKTIME_SETS_ERRNO
 # define MKTIME_SETS_ERRNO
+# undef HAVE_WORKING_64BIT_MKTIME
+# define HAVE_WORKING_64BIT_MKTIME 1
 #else
 
 typedef struct tm stm;
 #define R_tzname tzname
-# if defined(__CYGWIN__)
-extern __declspec(dllimport) char *tzname[2];
-# else
 extern char *tzname[2];
-# endif
 
 #endif
 
@@ -293,14 +299,12 @@ static Rboolean have_broken_mktime(void)
 }
 
 #ifndef HAVE_POSIX_LEAPSECONDS
-/* There have been 25 leapseconds: see .leap.seconds in R
- */
-static int n_leapseconds = 25;
-static const time_t leapseconds[] =
+static int n_leapseconds = 27; // 2017-01, sync with .leap.seconds in R (!)
+static const time_t leapseconds[] = // dput(unclass(.leap.seconds)) :
 {  78796800, 94694400,126230400,157766400,189302400,220924800,252460800,
   283996800,315532800,362793600,394329600,425865600,489024000,567993600,
   631152000,662688000,709948800,741484800,773020800,820454400,867715200,
-   915148800,1136073600,1230768000,1341100800};
+   915148800,1136073600,1230768000,1341100800,1435708800,1483228800};
 #endif
 
 static double guess_offset (stm *tm)
@@ -403,7 +407,7 @@ static double mktime0 (stm *tm, const int local)
     }
     if(!local) return mktime00(tm);
 
-/* OS X 10.9 gives -1 for dates prior to 1902, and ignores DST after 2037 */
+/* macOS 10.9 gives -1 for dates prior to 1902, and ignores DST after 2037 */
 #ifdef HAVE_WORKING_64BIT_MKTIME
     if(sizeof(time_t) == 8)
 	OK = !have_broken_mktime() || tm->tm_year >= 70;
@@ -436,7 +440,7 @@ static stm * localtime0(const double *tp, const int local, stm *ltm)
 	OK = !have_broken_mktime() || d > 0.;
     else
 #endif
-	OK = d < 2147483647.0 && 
+	OK = d < 2147483647.0 &&
 	    d > (have_broken_mktime() ? 0. : -2147483647.0);
     if(OK) {
 	t = (time_t) d;
@@ -453,7 +457,7 @@ static stm * localtime0(const double *tp, const int local, stm *ltm)
 	return local ? localtime(&t) : gmtime(&t);
 #endif
     }
-    
+
     /* internal substitute code.
        Like localtime, this returns a pointer to a static struct tm */
 
@@ -632,11 +636,11 @@ static void glibc_fix(stm *tm, int *invalid)
 
 
 static const char ltnames [][7] =
-{ "sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst", 
+{ "sec", "min", "hour", "mday", "mon", "year", "wday", "yday", "isdst",
   "zone",  "gmtoff"};
 
 
-static void 
+static void
 makelt(stm *tm, SEXP ans, R_xlen_t i, int valid, double frac_secs)
 {
     if(valid) {
@@ -658,10 +662,11 @@ makelt(stm *tm, SEXP ans, R_xlen_t i, int valid, double frac_secs)
 }
 
 
-             /* --------- R interfaces --------- */
+	     /* --------- R interfaces --------- */
 
 // We assume time zone names/abbreviations are ASCII, as all known ones are.
 
+// .Internal(as.POSIXlt(x, tz)) -- called only from  as.POSIXlt.POSIXct()
 SEXP attribute_hidden do_asPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
 {
     SEXP stz, x, ans, ansnames, klass, tzone;
@@ -726,7 +731,7 @@ SEXP attribute_hidden do_asPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
 	stm dummy, *ptm = &dummy;
 	double d = REAL(x)[i];
 	if(R_FINITE(d)) {
-	    ptm = localtime0(&d, 1 - isgmt, &dummy);
+	    ptm = localtime0(&d, !isgmt, &dummy);
 	    /* in theory localtime/gmtime always return a valid
 	       struct tm pointer, but Windows uses NULL for error
 	       conditions (like negative times). */
@@ -736,11 +741,11 @@ SEXP attribute_hidden do_asPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
 	if(!isgmt) {
 	    char *p = "";
 	    // or ptm->tm_zone
-	    if(valid && ptm->tm_isdst >= 0) 
+	    if(valid && ptm->tm_isdst >= 0)
 		p = R_tzname[ptm->tm_isdst];
 	    SET_STRING_ELT(VECTOR_ELT(ans, 9), i, mkChar(p));
 #ifdef HAVE_TM_GMTOFF
-	    INTEGER(VECTOR_ELT(ans, 10))[i] = 
+	    INTEGER(VECTOR_ELT(ans, 10))[i] =
 		valid ? (int)ptm->tm_gmtoff : NA_INTEGER;
 #endif
 	}
@@ -753,25 +758,25 @@ SEXP attribute_hidden do_asPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
     setAttrib(ans, install("tzone"), tzone);
     SEXP nm = getAttrib(x, R_NamesSymbol);
     if(nm != R_NilValue) setAttrib(VECTOR_ELT(ans, 5), R_NamesSymbol, nm);
-    UNPROTECT(6);
-
     if(settz) reset_tz(oldtz);
+    UNPROTECT(6);
     return ans;
 }
 
+// .Internal(as.POSIXct(x, tz)) -- called only from  as.POSIXct.POSIXlt()
 SEXP attribute_hidden do_asPOSIXct(SEXP call, SEXP op, SEXP args, SEXP env)
 {
     SEXP stz, x, ans;
     R_xlen_t n = 0, nlen[9];
     int isgmt = 0, settz = 0;
-    char oldtz[20] = "";
+    char oldtz[1001] = "";
     const char *tz = NULL;
     stm tm;
     double tmp;
 
     checkArity(op, args);
     PROTECT(x = duplicate(CAR(args))); /* coerced below */
-    if(!isVectorList(x) || LENGTH(x) < 9)
+    if(!isVectorList(x) || LENGTH(x) < 9) // must be 'POSIXlt'
 	error(_("invalid '%s' argument"), "x");
     if(!isString((stz = CADR(args))) || LENGTH(stz) != 1)
 	error(_("invalid '%s' value"), "tz");
@@ -801,9 +806,10 @@ SEXP attribute_hidden do_asPOSIXct(SEXP call, SEXP op, SEXP args, SEXP env)
     if(n > 0) {
 	for(int i = 0; i < 6; i++)
 	    if(nlen[i] == 0)
-		error(_("zero-length component in non-empty \"POSIXlt\" structure"));
+		error(_("zero-length component [[%d]] in non-empty \"POSIXlt\" structure"),
+		      i+1);
 	if(nlen[8] == 0)
-	    error(_("zero-length component in non-empty \"POSIXlt\" structure"));
+	    error(_("zero-length component [[%d]] in non-empty \"POSIXlt\" structure"), 9);
     }
     /* coerce fields to integer or real */
     SET_VECTOR_ELT(x, 0, coerceVector(VECTOR_ELT(x, 0), REALSXP));
@@ -823,21 +829,21 @@ SEXP attribute_hidden do_asPOSIXct(SEXP call, SEXP op, SEXP args, SEXP env)
 	tm.tm_mon   = INTEGER(VECTOR_ELT(x, 4))[i%nlen[4]];
 	tm.tm_year  = INTEGER(VECTOR_ELT(x, 5))[i%nlen[5]];
 	/* mktime ignores tm.tm_wday and tm.tm_yday */
-	tm.tm_isdst = isgmt ? 0:INTEGER(VECTOR_ELT(x, 8))[i%nlen[8]];
+	tm.tm_isdst = isgmt ? 0 : INTEGER(VECTOR_ELT(x, 8))[i%nlen[8]];
 	if(!R_FINITE(secs) || tm.tm_min == NA_INTEGER ||
 	   tm.tm_hour == NA_INTEGER || tm.tm_mday == NA_INTEGER ||
 	   tm.tm_mon == NA_INTEGER || tm.tm_year == NA_INTEGER)
 	    REAL(ans)[i] = NA_REAL;
 	else {
 	    errno = 0;
-	    tmp = mktime0(&tm, 1 - isgmt);
+	    tmp = mktime0(&tm, !isgmt);
 #ifdef MKTIME_SETS_ERRNO
 	    REAL(ans)[i] = errno ? NA_REAL : tmp + (secs - fsecs);
 #else
 	    REAL(ans)[i] = ((tmp == -1.)
 			    /* avoid silly gotcha at epoch minus one sec */
 			    && (tm.tm_sec != 59)
-			    && ((tm.tm_sec = 58), (mktime0(&tm, 1 - isgmt) != -2.))
+			    && ((tm.tm_sec = 58), (mktime0(&tm, !isgmt) != -2.))
 			    ) ?
 	      NA_REAL : tmp + (secs - fsecs);
 #endif
@@ -852,24 +858,23 @@ SEXP attribute_hidden do_asPOSIXct(SEXP call, SEXP op, SEXP args, SEXP env)
 
 SEXP attribute_hidden do_formatPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
 {
-    SEXP x, sformat, ans, tz;
-    R_xlen_t n = 0, m, N, nlen[9];
-    int UseTZ, settz = 0;
+    int settz = 0;
     char buff[300];
-    char oldtz[20] = "";
+    char oldtz[1001] = "";
     stm tm;
 
     checkArity(op, args);
-    PROTECT(x = duplicate(CAR(args))); /* coerced below */
+    SEXP x = PROTECT(duplicate(CAR(args))); /* coerced below */
     if(!isVectorList(x) || LENGTH(x) < 9)
 	error(_("invalid '%s' argument"), "x");
+    SEXP sformat;
     if(!isString((sformat = CADR(args))) || XLENGTH(sformat) == 0)
 	error(_("invalid '%s' argument"), "format");
-    m = XLENGTH(sformat);
-    UseTZ = asLogical(CADDR(args));
+    R_xlen_t m = XLENGTH(sformat);
+    int UseTZ = asLogical(CADDR(args));
     if(UseTZ == NA_LOGICAL)
 	error(_("invalid '%s' argument"), "usetz");
-    tz = getAttrib(x, install("tzone"));
+    SEXP tz = getAttrib(x, install("tzone"));
 
     const char *tz1;
     if (!isNull(tz) && strlen(tz1 = CHAR(STRING_ELT(tz, 0)))) {
@@ -883,27 +888,39 @@ SEXP attribute_hidden do_formatPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
 	if(needTZ) settz = set_tz(tz1, oldtz);
     }
 
-    /* workaround for glibc/FreeBSD/MacOS X strftime: they have
+    /* workaround for glibc/FreeBSD/macOS strftime: they have
        non-POSIX/C99 time zone components
      */
     memset(&tm, 0, sizeof(tm));
 
-    /* coerce fields to integer or real, find length of longest one */
-    for(int i = 0; i < 9; i++) {
+    /* coerce fields, find length of longest one */
+    R_xlen_t n = 0, nlen[11];
+    int nn = imin2(LENGTH(x), 11);
+    for(int i = 0; i < nn; i++) {
 	nlen[i] = XLENGTH(VECTOR_ELT(x, i));
 	if(nlen[i] > n) n = nlen[i];
-	SET_VECTOR_ELT(x, i, coerceVector(VECTOR_ELT(x, i),
-					  i > 0 ? INTSXP : REALSXP));
+	if(i != 9) // real for 'sec', the first; integer for the rest:
+	    SET_VECTOR_ELT(x, i, coerceVector(VECTOR_ELT(x, i),
+					      i > 0 ? INTSXP : REALSXP));
     }
     if(n > 0) {
-	for(int i = 0; i < 9; i++)
+	for(int i = 0; i < nn; i++)
 	    if(nlen[i] == 0)
-		error(_("zero-length component in non-empty \"POSIXlt\" structure"));
+		error(_("zero-length component [[%d]] in non-empty \"POSIXlt\" structure"),
+		      i+1);
     }
-    if(n > 0) N = (m > n) ? m:n; else N = 0;
-    PROTECT(ans = allocVector(STRSXP, N));
+    R_xlen_t N = (n > 0) ? ((m > n) ? m : n) : 0;
+    SEXP ans = PROTECT(allocVector(STRSXP, N));
     char tm_zone[20];
-    Rboolean have_zone = LENGTH(x) >= 10 && LENGTH(VECTOR_ELT(x, 9)) == n;
+#ifdef HAVE_TM_GMTOFF
+    Rboolean have_zone = LENGTH(x) >= 11;// and components w/ length >= 1
+#else
+    Rboolean have_zone = LENGTH(x) >= 10;
+#endif
+    if(have_zone && !isString(VECTOR_ELT(x, 9)))
+	error(_("invalid component [[10]] in \"POSIXlt\" should be 'zone'"));
+    if(!have_zone && LENGTH(x) > 9) // rather even error ?
+	warning(_("More than 9 list components in \"POSIXlt\" without timezone"));
     for(R_xlen_t i = 0; i < N; i++) {
 	double secs = REAL(VECTOR_ELT(x, 0))[i%nlen[0]], fsecs = floor(secs);
 	// avoid (int) NAN
@@ -916,9 +933,9 @@ SEXP attribute_hidden do_formatPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
 	tm.tm_wday  = INTEGER(VECTOR_ELT(x, 6))[i%nlen[6]];
 	tm.tm_yday  = INTEGER(VECTOR_ELT(x, 7))[i%nlen[7]];
 	tm.tm_isdst = INTEGER(VECTOR_ELT(x, 8))[i%nlen[8]];
-	if(have_zone) {
-	    strncpy(tm_zone, CHAR(STRING_ELT(VECTOR_ELT(x, 9), i)), 20);
-            tm_zone[20 - 1] = '\0';
+	if(have_zone) { // not "UTC", e.g.
+	    strncpy(tm_zone, CHAR(STRING_ELT(VECTOR_ELT(x, 9), i%nlen[9])), 20 - 1);
+	    tm_zone[20 - 1] = '\0';
 #ifdef HAVE_TM_ZONE
 	    tm.tm_zone = tm_zone;
 #elif defined USE_INTERNAL_MKTIME
@@ -928,7 +945,7 @@ SEXP attribute_hidden do_formatPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
 	    if(tm.tm_isdst >= 0) tzname[tm.tm_isdst] = tm_zone;
 #endif
 #ifdef HAVE_TM_GMTOFF
-	    int tmp = INTEGER(VECTOR_ELT(x, 10))[i%n];
+	    int tmp = INTEGER(VECTOR_ELT(x, 10))[i%nlen[10]];
 	    tm.tm_gmtoff = (tmp == NA_INTEGER) ? 0 : tmp;
 #endif
 	}
@@ -936,89 +953,95 @@ SEXP attribute_hidden do_formatPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
 	   tm.tm_hour == NA_INTEGER || tm.tm_mday == NA_INTEGER ||
 	   tm.tm_mon == NA_INTEGER || tm.tm_year == NA_INTEGER) {
 	    SET_STRING_ELT(ans, i, NA_STRING);
+	} else if(validate_tm(&tm) < 0) {
+	    SET_STRING_ELT(ans, i, NA_STRING);
 	} else {
-	    if(validate_tm(&tm) < 0) SET_STRING_ELT(ans, i, NA_STRING);
-	    else {
-		const char *q = translateChar(STRING_ELT(sformat, i%m));
-		int n = (int) strlen(q) + 50;
-		char buf2[n];
-		const char *p;
+	    const char *q = translateChar(STRING_ELT(sformat, i%m));
+	    int nn = (int) strlen(q) + 50;
+	    char buf2[nn];
+	    const char *p;
 #ifdef OLD_Win32
-		/* We want to override Windows' TZ names */
-		p = strstr(q, "%Z");
-		if (p) {
-		    memset(buf2, 0, n);
-		    strncpy(buf2, q, p - q);
-		    if(have_zone) 
-			strcat(buf2, tm_zone);
-		    else
-			strcat(buf2, tm.tm_isdst > 0 ? R_tzname[1] : R_tzname[0]);
-		    strcat(buf2, p+2);
-		} else
+	    /* We want to override Windows' TZ names */
+	    p = strstr(q, "%Z");
+	    if (p) {
+		memset(buf2, 0, nn);
+		strncpy(buf2, q, p - q);
+		if(have_zone)
+		    strcat(buf2, tm_zone);
+		else
+		    strcat(buf2, tm.tm_isdst > 0 ? R_tzname[1] : R_tzname[0]);
+		strcat(buf2, p+2);
+	    } else
 #endif
-		    strcpy(buf2, q);
+		strcpy(buf2, q);
 
-		p = strstr(q, "%OS");
-		if(p) {
-		    /* FIXME some of this should be outside the loop */
-		    int ns, nused = 4;
-		    char *p2 = strstr(buf2, "%OS");
-		    *p2 = '\0';
-		    ns = *(p+3) - '0';
-		    if(ns < 0 || ns > 9) { /* not a digit */
-			ns = asInteger(GetOption1(install("digits.secs")));
-			if(ns == NA_INTEGER) ns = 0;
-			nused = 3;
-		    }
-		    if(ns > 6) ns = 6;
-		    if(ns > 0) {
-			/* truncate to avoid nuisances such as PR#14579 */
-			double s = secs, t = Rexp10((double) ns);
-			s = ((int) (s*t))/t;
-			sprintf(p2, "%0*.*f", ns+3, ns, s);
-			strcat(buf2, p+nused);
-		    } else {
-			strcat(p2, "%S");
-			strcat(buf2, p+nused);
-		    }
+	    p = strstr(q, "%OS");
+	    if(p) {
+		/* FIXME some of this should be outside the loop */
+		int ns, nused = 4;
+		char *p2 = strstr(buf2, "%OS");
+		*p2 = '\0';
+		ns = *(p+3) - '0';
+		if(ns < 0 || ns > 9) { /* not a digit */
+		    ns = asInteger(GetOption1(install("digits.secs")));
+		    if(ns == NA_INTEGER) ns = 0;
+		    nused = 3;
 		}
-#ifdef USE_INTERNAL_MKTIME
-		R_strftime(buff, 256, buf2, &tm);
-#else
-		strftime(buff, 256, buf2, &tm);
-#endif
-		if(UseTZ) {
-		    if(LENGTH(x) >= 10) {
-			const char *p = CHAR(STRING_ELT(VECTOR_ELT(x, 9), i));
-			if(strlen(p)) {strcat(buff, " "); strcat(buff, p);}
-		    } else if(!isNull(tz)) {
-			int ii = 0;
-			if(LENGTH(tz) == 3) {
-			    if(tm.tm_isdst > 0) ii = 2;
-			    else if(tm.tm_isdst == 0) ii = 1;
-			    else ii = 0; /* Use base timezone name */
-			}
-			const char *p = CHAR(STRING_ELT(tz, ii));
-			if(strlen(p)) {strcat(buff, " "); strcat(buff, p);}
-		    }
+		if(ns > 6) ns = 6;
+		if(ns > 0) {
+		    /* truncate to avoid nuisances such as PR#14579 */
+		    double s = secs, t = Rexp10((double) ns);
+		    s = ((int) (s*t))/t;
+		    sprintf(p2, "%0*.*f", ns+3, ns, s);
+		    strcat(buf2, p+nused);
+		} else {
+		    strcat(p2, "%S");
+		    strcat(buf2, p+nused);
 		}
-		SET_STRING_ELT(ans, i, mkChar(buff));
 	    }
+	    // The overflow behaviour is not determined by C99.
+	    // We assume truncation, and ensure termination.
+#ifdef USE_INTERNAL_MKTIME
+	    R_strftime(buff, 256, buf2, &tm);
+#else
+	    strftime(buff, 256, buf2, &tm);
+#endif
+	    buff[256] = '\0';
+	    mbcsTruncateToValid(buff);
+	    // Now assume tzone abbreviated name is < 40 bytes,
+	    // but they are currently 3 or 4 bytes.
+	    if(UseTZ) {
+		if(have_zone) {
+		    const char *p = CHAR(STRING_ELT(VECTOR_ELT(x, 9), i%nlen[9]));
+		    if(strlen(p)) {strcat(buff, " "); strcat(buff, p);}
+		} else if(!isNull(tz)) {
+		    int ii = 0;
+		    if(LENGTH(tz) == 3) {
+			if(tm.tm_isdst > 0) ii = 2;
+			else if(tm.tm_isdst == 0) ii = 1;
+			else ii = 0; /* Use base timezone name */
+		    }
+		    const char *p = CHAR(STRING_ELT(tz, ii));
+		    if(strlen(p)) {strcat(buff, " "); strcat(buff, p);}
+		}
+	    }
+	    SET_STRING_ELT(ans, i, mkChar(buff));
 	}
     }
-    UNPROTECT(2);
     if(settz) reset_tz(oldtz);
+    UNPROTECT(2);
     return ans;
 }
 
 
+// .Internal(strptime(as.character(x), format, tz))
 SEXP attribute_hidden do_strptime(SEXP call, SEXP op, SEXP args, SEXP env)
 {
     SEXP x, sformat, ans, ansnames, klass, stz, tzone = R_NilValue;
     int invalid, isgmt = 0, settz = 0, offset;
     stm tm, tm2, *ptm = &tm;
     const char *tz = NULL;
-    char oldtz[20] = "";
+    char oldtz[1001] = "";
     double psecs = 0.0;
     R_xlen_t n, m, N;
 
@@ -1026,7 +1049,7 @@ SEXP attribute_hidden do_strptime(SEXP call, SEXP op, SEXP args, SEXP env)
     if(!isString((x = CAR(args))))
 	error(_("invalid '%s' argument"), "x");
     if(!isString((sformat = CADR(args))) || XLENGTH(sformat) == 0)
-	error(_("invalid '%s' argument"), "x");
+	error(_("invalid '%s' argument"), "format");
     if(!isString((stz = CADDR(args))) || LENGTH(stz) != 1)
 	error(_("invalid '%s' value"), "tz");
     tz = CHAR(STRING_ELT(stz, 0));
@@ -1056,7 +1079,7 @@ SEXP attribute_hidden do_strptime(SEXP call, SEXP op, SEXP args, SEXP env)
 	SET_STRING_ELT(tzone, 0, mkChar(tz));
 	SET_STRING_ELT(tzone, 1, mkChar(R_tzname[0]));
 	SET_STRING_ELT(tzone, 2, mkChar(R_tzname[1]));
-	
+
     } else PROTECT(tzone); // for balance
 
     n = XLENGTH(x); m = XLENGTH(sformat);
@@ -1071,9 +1094,9 @@ SEXP attribute_hidden do_strptime(SEXP call, SEXP op, SEXP args, SEXP env)
     for(int i = 0; i < 9; i++)
 	SET_VECTOR_ELT(ans, i, allocVector(i > 0 ? INTSXP : REALSXP, N));
     if(!isgmt) {
-	SET_VECTOR_ELT(ans, 9, allocVector(STRSXP, n));
+	SET_VECTOR_ELT(ans, 9, allocVector(STRSXP, N));
 #ifdef HAVE_TM_GMTOFF
-	SET_VECTOR_ELT(ans, 10, allocVector(INTSXP, n));
+	SET_VECTOR_ELT(ans, 10, allocVector(INTSXP, N));
 #endif
     }
 
@@ -1123,7 +1146,7 @@ SEXP attribute_hidden do_strptime(SEXP call, SEXP op, SEXP args, SEXP env)
 		/* we do want to set wday, yday, isdst, but not to
 		   adjust structure at DST boundaries */
 		memcpy(&tm2, &tm, sizeof(stm));
-		mktime0(&tm2, 1-isgmt); /* set wday, yday, isdst */
+		mktime0(&tm2, !isgmt); /* set wday, yday, isdst */
 		tm.tm_wday = tm2.tm_wday;
 		tm.tm_yday = tm2.tm_yday;
 		tm.tm_isdst = isgmt ? 0: tm2.tm_isdst;
@@ -1142,7 +1165,7 @@ SEXP attribute_hidden do_strptime(SEXP call, SEXP op, SEXP args, SEXP env)
 	    }
 	    SET_STRING_ELT(VECTOR_ELT(ans, 9), i, mkChar(p));
 #ifdef HAVE_TM_GMTOFF
-	    INTEGER(VECTOR_ELT(ans, 10))[i] = 
+	    INTEGER(VECTOR_ELT(ans, 10))[i] =
 		invalid ? NA_INTEGER : (int)tm.tm_gmtoff;
 #endif
 	}
@@ -1213,7 +1236,8 @@ SEXP attribute_hidden do_D2POSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
     SET_STRING_ELT(klass, 0, mkChar("POSIXlt"));
     SET_STRING_ELT(klass, 1, mkChar("POSIXt"));
     classgets(ans, klass);
-    setAttrib(ans, install("tzone"), mkString("UTC"));
+    SEXP s_tzone = install("tzone");
+    setAttrib(ans, s_tzone, mkString("UTC"));
     SEXP nm = getAttrib(x, R_NamesSymbol);
     if(nm != R_NilValue) setAttrib(VECTOR_ELT(ans, 5), R_NamesSymbol, nm);
     UNPROTECT(4);
@@ -1238,9 +1262,10 @@ SEXP attribute_hidden do_POSIXlt2D(SEXP call, SEXP op, SEXP args, SEXP env)
     if(n > 0) {
 	for(int i = 3; i < 6; i++)
 	    if(nlen[i] == 0)
-		error(_("zero-length component in non-empty \"POSIXlt\" structure"));
+		error(_("zero-length component [[%d]] in non-empty \"POSIXlt\" structure"),
+		      i+1);
 	if(nlen[8] == 0)
-	    error(_("zero-length component in non-empty \"POSIXlt\" structure"));
+	    error(_("zero-length component [[%d]] in non-empty \"POSIXlt\" structure"), 9);
     }
     /* coerce relevant fields to integer */
     for(int i = 3; i < 6; i++)
